// Copyright 2008 Michiaki Hamada

#ifndef __H_INCLUDE_LAMBDA_CALCULATOR_HH
#define __H_INCLUDE_LAMBDA_CALCULATOR_HH

// These pointers are 1-based!
double calculate_lambda( const double** mat_b, int alpha_size,
                         double* p, double* q );

#endif // __H_INCLUDE_LAMBDA_CALCULATOR_HH
